/***************************************************************************/
/**                                                                       **/
/**                 The ORIC 48K emulator.UNIX version                    **/
/**                                                                       **/
/**                Written by JF FABRE (fabre@supaero.fr)                 **/
/**                                                                       **/
/**                This program may be freely distibuted                  **/
/**                                                                       **/
/**             as long as this header is included unmodified             **/
/**                                                                       **/
/***************************************************************************/

/* 6502 Emulation routines */

#include "protos.h"
#include "macros.h"

extern unsigned short int PC;
extern unsigned char S,X,A,Y,P;
extern unsigned char *M;
extern unsigned char NC;

extern SND sons;
extern ecran disp;
extern Monitor_struct mon;
extern comm_struct cs;

int PHA();
int RTS();
void LDA();

int ReadInstruction()


{
return (**(PFONCT[M[PC]]))();
}

/*********************/

void Empiler(c)

unsigned char c;

{
M[(unsigned short int)S+STACK_BEGIN]=c;
S--;
}

/*********************/

unsigned char Depiler()

{
return(M[(unsigned short int)(++S)+STACK_BEGIN]);
}

/*** code du Trap ***/

int TrapOPCode()

{
register unsigned char wrk,i;

switch(M[PC])

 {
 case 0x17 : /* breakpoint signaled */
             return BREAKPOINT;

 case 0x27 : /* OUTBYT or RDBYTE or PRINTER */
 
             if (PC==J_RDBYTE) return PATCH_RDBYTE;
             if (PC==J_OUTBYT) return PATCH_OUTBYT;
             if (PC==J_SYNC)   return   PATCH_SYNC;
             if (PC==J_OUTCHR) return   PATCH_PRINT;
             return ILLEGAL_INSTRUCTION;

 case 0x37 : /* HIRES ou TEXT */
 
   switch(PC)
      {
      case J_RETURN_INIT :
                mon.dispmode=TEXT;        
                RefreshDisplay();
                RewindTapeLoad();     
                     /** remet le fichier cassette a zero s'il existe **/
                PC=0xECCC;
                return NEXT_INSTRUCTION;
      case J_RETURN_HIRES :
                RefreshDisplay();
                return RTS();
                
      default : return ILLEGAL_INSTRUCTION;
      }
                                  
              
 case 0xFF : if (!mon.sound) return RTS();
     switch (PC)
      {
      case J_ZAP     : write(cs.fd_pere[1],"Z",1);break;
      case J_PING    : write(cs.fd_pere[1],"P",1);break;
      case J_EXPLODE : write(cs.fd_pere[1],"E",1);break;
      case J_SHOOT   : write(cs.fd_pere[1],"S",1);break;
      case J_CONTBP  : write(cs.fd_pere[1],"C",1);break;
      case J_KBBEEP  : write(cs.fd_pere[1],"K",1);break;
      case J_MUSIC   : return RTS();
      case J_SOUND   : return RTS();
      case J_PLAY    : return RTS();
      default        : return ILLEGAL_INSTRUCTION;
      }
      return RTS();
      
 }
return NEXT_INSTRUCTION;
}

/*** code  fock ***/

int IllegalInstruction()

{
return ILLEGAL_INSTRUCTION;
}

/*** ADC generique ***/

void ADC(B)

register unsigned char B;
{
register unsigned char AC,BC,RET;
unsigned char AD,BD;
register short int INTER;

RET=P&C_BIT;


  if (P&D_BIT)
  {
  /* Mode decimal */
  
  P|=0x03; /* posit ZC */

  AC=A&0xF;
  BC=B&0xF;
  AD=A&0xF0;
  BD=B&0xF0;
  
  if ((AC+BC+RET)>9) {INTER=AC+BC+RET-10;AD+=0x10;} else INTER=AC+BC+RET;
  if (((int)(AD+BD))>0x90) {A=INTER+(int)AD+(int)BD-0xA0;} 
                        else {A=INTER+AD+BD;P^=0x1;}
  
  if (A) P^=0x2;
  return;
  }


  /** Mode binaire **/ 

  P|=0xC3; /* posit CVZN */
   
  if (((unsigned short int)A+(unsigned short int)B+(unsigned short int)RET)>255) 
       {AC=A+B+RET;} else {AC=A+B+RET;P^=0x1;} 
       /** positionnement de C=0 si retenue **/

  INTER=(short int)((char)A+(char)B+RET);
  
  if ((INTER>-128)&&(INTER<127)) 
             P^=0x40; /** overflow V si nombre non codable en signe  **/

 
  A=AC;
  if (AC)       P^=0x02; /** positionnement du bit Z=0 si A!=0 **/
  if (AC<=127)  P^=0x80; /** positionnement du bit N=0 si positif ***/
  
}

/*** SBC generique ***/

void SBC(B)

register unsigned char B;
{
register unsigned char AC,BC,RET;
unsigned char AD,BD;
register short int INTER;

RET=1-(P&C_BIT);

P|=0xc3; /* posit CVZN */

if (P&D_BIT)
  {
  /* Mode decimal */

  P|=Z_BIT|C_BIT; /* posit ZC */

  AC=A&0xF;
  BC=B&0xF;
  AD=A&0xF0;
  BD=B&0xF0;
  
  if ((AC+BC+RET)>9) {INTER=AC-BC-RET+10;AD-=0x10;} else INTER=AC-BC-RET;
  if (((int)(AD-BD))>0x90) {A=INTER-(int)AD-(int)BD+0xA0;} 
                        else {A=INTER-AD-BD;P^=C_BIT;}
  
  if (A) P^=Z_BIT;
  return;
  
  }
  
  INTER=A-B-RET;
  if (INTER<0) {AC=INTER;P^=0x1;} else {AC=INTER;} /** positionnement de C=0 si retenue **/

  INTER=(short int)((char)A-(char)B-RET);
  
  if ((INTER>-128)&&(INTER<127)) 
             P^=V_BIT; /** overflow V si nombre non codable en signe  **/

  
if (AC)       P^=Z_BIT; /** positionnement du bit Z=0 si A!=0 **/
if (AC<=127)  P^=N_BIT; /** positionnement du bit N=0 si positif ***/

A=AC;
}

/*** BIT generique ***/

void BIT(B)

register unsigned char B;
{
if (A&B) P&=0xFF-Z_BIT; /** positionnement du bit Z si A=B **/
    else P|=Z_BIT;
    
P=(P&0x3F)|(B&0xC0); /** gaule les bits 6 et 7 de B et les positionne dans P **/  
}

/*** LDA generique ***/

void LDA(B)

register unsigned char B;
{
LDx(A);
}

/*** LDX generique ***/

void LDX(B)

register unsigned char B;
{
LDx(X);
}

/*** LDY generique ***/

void LDY(B)

register unsigned char B;
{
LDx(Y);
}

/*** CMP generique ***/

void CMP(B)

register unsigned char B;
{
P|=0x83; /* posit NZC */
if (A==B)  {P^=N_BIT;return;}
if (A<B)   P^=C_BIT; 
if ((unsigned char)(A-B)<=127) P^=N_BIT;
  
P^=Z_BIT;
}

/*** CPX generique ***/

void CPX(B)

register unsigned char B;
{
P|=0x83; /* posit NZC */
if (X==B) {P^=0x80;return;}
if (X<B)   P^=0x1; 
if ((unsigned char)(X-B)<=127) P^=0x80;  
P^=2;
}

/*** CPY generique ***/

void CPY(B)

register unsigned char B;
{
P|=0x83; /* posit NZC */
if (Y==B) {P^=0x80;return;}
if (Y<B)   P^=0x1; 

if ((unsigned char)(Y-B)<=127) P^=0x80;  
P^=2;
}

/*** ROL generique ***/

unsigned char ROL(valeur)

register unsigned char valeur;

{
register unsigned char C,R;


C=((((unsigned short int)valeur)&0x80)>>7);
R=((((short int)valeur))<<1)|(P&0x1);
P&=0x7C; /* masque les bits NZC */
P|=(short int)C;
if (R>127) P|=0x80;       /* test negatif */
if (!R) P|=0x2;           /* test nul */

return R;
}

/*** ROR generique ***/

unsigned char ROR(valeur)

register unsigned char valeur;

{
register unsigned char C,R;


C=valeur&0x1;
R=((((short int)valeur)>>1)|((P&0x1)<<7));
P&=0x7C; /* masque les bits NZC */
P|=(short int)C;
if (R>127) P|=0x80;          /* test negatif */
if (!R) P|=0x2;              /* test nul */

return R;
}

/*** AND generique ***/

void AND(B)

register unsigned char B;

{
A&=B;    /* Logical AND between A and B */
P|=0x82;
if (A) P^=Z_BIT;
if (A<=127) P^=N_BIT;
}

/*** EOR generique ***/

void EOR(B)

register unsigned char B;

{
A^=B;    /* Logical EOR between A and B */
P|=0x82;
if (A) P^=0x2;
if (A<=127) P^=0x80;
}

/*** ORA generique ***/

void ORA(B)

register unsigned char B;

{
A|=B;    /* Logical ORA between A and B */
P|=0x82;
if (A) P^=0x2;
if (A<=127) P^=0x80;
}

/*** ASL generique ***/

unsigned char ASL(B)

register unsigned char B;

{
register unsigned char R;

if (B&0x80) P|=0x1; else P&=0xFE; /* retenue si bit 7 =1 */
R=B<<1;
if (R) P&=0xFD; else P|=0x2; /* test du Zero */
if (R>127) P|=0x80; else P&=0x7F; /* test Negatif */
return R;
}

/*** LSR generique ***/

unsigned char LSR(B)

register unsigned char B;

{
register unsigned char R;

if (B&0x01) P|=0x1; else P&=0xFE; /* retenue si bit 0 =1 */
R=((short int)B)>>1;
if (R) P&=0xFD; else P|=0x2; /* test du Zero */
P&=0x7F; /* forcement N=0 */
return R;
}


/*** INC generique ***/

void INC(B)

register unsigned short int B;
{
register unsigned char DTA;


DTA=M[B]+1;

Place(DTA,B);
TestNZ(DTA);

}

/*** DEC generique ***/

void DEC(B)

register unsigned short int B;
{
register unsigned char DTA;

DTA=M[B]-1;

Place(DTA,B);
TestNZ(DTA);

}

/*****************************************/
/************** codes OP *****************/
/*****************************************/

/*** Interrupt Request ***/

void IRQ()

{
if (P&I_BIT) return; /** si interruptions masquees ne fait rien **/

Empiler((unsigned char)(PC>>8));
Empiler((unsigned char)(PC&0xFF));
Empiler(P);  /** sauvegarde de P dans la pile **/

PC=M[0xFFFE]+((unsigned short int)M[0xFFFF]<<8);  /** gestion de l'interruption **/

}


/*** code 0 : BRK ***/

int BRK()

{
P|=B_BIT;       /** B=1 **/
PC+=2;
Empiler((unsigned char)(PC>>8));
Empiler((unsigned char)(PC&0xFF));
Empiler(P);  /** sauvegarde de P dans la pile **/

NC=7; /* 7 cycles horloge */

PC=M[0xFFFE]+((unsigned short int)M[0xFFFF]<<8);  /** gestion de l'interruption **/
return NEXT_INSTRUCTION;      /** instruction correctement decodee **/
}

/*** code 24 : BIT Zero Page **/

int BIT_ZRP()

{
NC=3;
A_ZRP(BIT);
}

/*** code 2C : BIT Absolu ***/

int BIT_ABS()

{
NC=4;
A_ABS(BIT);
}

/*** code 20 : JSR **/

int JSR()

{
PC++;
Empiler((unsigned char)((PC+1)>>8));
Empiler((unsigned char)((PC+1)&0xFF));
NC=6;

PC=M[PC]+(M[PC+1]<<8);
return NEXT_INSTRUCTION;
}

/*** code 4C : JMP_IMM **/

int JMP_IMM()

{
PC=M[PC+1]+(M[PC+2]<<8);
NC=3;

return NEXT_INSTRUCTION;
}

/*** code 6C : JMP_IND **/

int JMP_IND()

{
register unsigned short int IN;

IN=M[PC+1]+(M[PC+2]<<8);
PC=M[IN]+(M[IN+1]<<8);
NC=5;

return NEXT_INSTRUCTION;
}

/*** code 40 : RTI **/

int RTI()

{
P=Depiler();
PC=Depiler()+(Depiler()<<8); /** page 1 : pile **/
NC=6;
return NEXT_INSTRUCTION;
}

/*** code 60 : RTS **/

int RTS()

{
PC=Depiler()+(Depiler()<<8)+1; /** page 1 : pile **/
NC=6;
return NEXT_INSTRUCTION;
}

/***************************************************************/
/*******************    ADC     ********************************/
/***************************************************************/


/*** code 69 : ADC immediat ***/

int ADC_IMM()

{
A_IMM(ADC);
}

/*** code 65 : ADC zero page ***/

int ADC_ZRP()

{
A_ZRP(ADC);
}

/*** code 75 : ADC zero page X ***/

int ADC_ZRPX()

{
A_ZRPX(ADC);
}
/*** code 61 : ADC_INDX ***/

int ADC_INDX()

{
A_INDX(ADC);
}

/*** code 71 : ADC_INDY ***/

int ADC_INDY()

{
A_INDY(ADC);
}

/*** code 7D : ADC_ABS ***/

int ADC_ABS()

{
A_ABS(ADC);
}

/*** code 79 : ADC_ABSX ***/

int ADC_ABSX()

{
A_ABSX(ADC);
}

/*** code 79 : ADC_ABSY ***/

int ADC_ABSY()

{
A_ABSY(ADC);
}

/***************************************************************/
/*******************    SBC     ********************************/
/***************************************************************/

/*** code 69 : SBC immediat ***/

int SBC_IMM()

{
A_IMM(SBC);
}

/*** code 65 : SBC zero page ***/

int SBC_ZRP()

{
A_ZRP(SBC);
}

/*** code 75 : SBC zero page X ***/

int SBC_ZRPX()

{
A_ZRPX(SBC);
}
/*** code 61 : SBC_INDX ***/

int SBC_INDX()

{
A_INDX(SBC);
}

/*** code 71 : SBC_INDY ***/

int SBC_INDY()

{
A_INDY(SBC);
}

/*** code 7D : SBC_ABS ***/

int SBC_ABS()

{
A_ABS(SBC);
}

/*** code 79 : SBC_ABSX ***/

int SBC_ABSX()

{
A_ABSX(SBC);
}

/*** code 79 : SBC_ABSY ***/

int SBC_ABSY()

{
A_ABSY(SBC);
}

/***************************************************************/
/*******************    LDA     ********************************/
/***************************************************************/

/*** code A5 : LDA_ZRP ***/

int LDA_ZRP()

{
A_ZRP(LDA);
}

/*** code A9 : LDA_IMM ***/

int LDA_IMM()

{
A_IMM(LDA);
}

/*** code B5 : LDA_ZRPX ***/

int LDA_ZRPX()

{
A_ZRPX(LDA);
}

/*** code A1 : LDA_INDX ***/

int LDA_INDX()

{
A_INDX(LDA);
}

/*** code B1 : LDA_INDY ***/

int LDA_INDY()

{
A_INDY(LDA);
}

/*** code AD : LDA_ABS ***/

int LDA_ABS()

{
A_ABS(LDA);
}

/*** code BD : LDA_ABSX ***/

int LDA_ABSX()

{
A_ABSX(LDA);
}

/*** code B9 : LDA_ABSY ***/

int LDA_ABSY()

{
A_ABSY(LDA);
}

/***************************************************************/
/*******************    AND     ********************************/
/***************************************************************/

/*** code A5 : AND_ZRP ***/

int AND_ZRP()

{
A_ZRP(AND);
}

/*** code A9 : AND_IMM ***/

int AND_IMM()

{
A_IMM(AND);
}

/*** code B5 : AND_ZRPX ***/

int AND_ZRPX()

{
A_ZRPX(AND);
}

/*** code A1 : AND_INDX ***/

int AND_INDX()

{
A_INDX(AND);
}

/*** code B1 : AND_INDY ***/

int AND_INDY()

{
A_INDY(AND);
}

/*** code AD : AND_ABS ***/

int AND_ABS()

{
A_ABS(AND);
}

/*** code BD : AND_ABSX ***/

int AND_ABSX()

{
A_ABSX(AND);
}

/*** code B9 : AND_ABSY ***/

int AND_ABSY()

{
A_ABSY(AND);
}


/***************************************************************/
/*******************    EOR     ********************************/
/***************************************************************/

/*** code 45 : EOR_ZRP ***/

int EOR_ZRP()

{
A_ZRP(EOR);
}

/*** code 49 : EOR_IMM ***/

int EOR_IMM()

{
A_IMM(EOR);
}

/*** code 55 : EOR_ZRPX ***/

int EOR_ZRPX()

{
A_ZRPX(EOR);
}

/*** code 41 : EOR_INDX ***/

int EOR_INDX()

{
A_INDX(EOR);
}

/*** code 51 : EOR_INDY ***/

int EOR_INDY()

{
A_INDY(EOR);
}

/*** code 4D : EOR_ABS ***/

int EOR_ABS()

{
A_ABS(EOR);
}

/*** code 5D : EOR_ABSX ***/

int EOR_ABSX()

{
A_ABSX(EOR);
}

/*** code 59 : EOR_ABSY ***/

int EOR_ABSY()

{
A_ABSY(EOR);
}


/***************************************************************/
/*******************    ORA     ********************************/
/***************************************************************/

/*** code 05 : ORA_ZRP ***/

int ORA_ZRP()

{
A_ZRP(ORA);
}

/*** code 09 : ORA_IMM ***/

int ORA_IMM()

{
A_IMM(ORA);
}

/*** code 15 : ORA_ZRPX ***/

int ORA_ZRPX()

{
A_ZRPX(ORA);
}

/*** code 01 : ORA_INDX ***/

int ORA_INDX()

{
A_INDX(ORA);
}

/*** code 11 : ORA_INDY ***/

int ORA_INDY()

{
A_INDY(ORA);
}

/*** code 0D : ORA_ABS ***/

int ORA_ABS()

{
A_ABS(ORA);
}

/*** code 1D : ORA_ABSX ***/

int ORA_ABSX()

{
A_ABSX(ORA);
}

/*** code 19 : ORA_ABSY ***/

int ORA_ABSY()

{
A_ABSY(ORA);
}

/***************************************************************/
/*******************    LDX     ********************************/
/***************************************************************/

/*** code A6 : LDX_ZRP ***/

int LDX_ZRP()

{
A_ZRP(LDX);
}

/*** code A2 : LDX_IMM ***/

int LDX_IMM()

{
A_IMM(LDX);
}

/*** code B6 : LDX_ZRPY ***/

int LDX_ZRPY()

{
A_ZRPY(LDX);
}

/*** code AE : LDX_ABS ***/

int LDX_ABS()

{
A_ABS(LDX);
}


/*** code BE : LDX_ABSY ***/

int LDX_ABSY()

{
A_ABSY(LDX);
}


/***************************************************************/
/*******************    LDY     ********************************/
/***************************************************************/

/*** code A4 : LDY_ZRP ***/

int LDY_ZRP()

{
A_ZRP(LDY);
}

/*** code A0 : LDY_IMM ***/

int LDY_IMM()

{
A_IMM(LDY);
}

/*** code B4 : LDY_ZRPX ***/

int LDY_ZRPX()

{
A_ZRPX(LDY);
}

/*** code AE : LDY_ABS ***/

int LDY_ABS()

{
A_ABS(LDY);
}


/*** code BC : LDY_ABSX ***/

int LDY_ABSX()

{
A_ABSX(LDY);
}

/***************************************************************/
/*******************    CMP     ********************************/
/***************************************************************/

/*** code A5 : CMP_ZRP ***/

int CMP_ZRP()

{
A_ZRP(CMP);
}

/*** code A9 : CMP_IMM ***/

int CMP_IMM()

{
A_IMM(CMP);
}

/*** code B5 : CMP_ZRPX ***/

int CMP_ZRPX()

{
A_ZRPX(CMP);
}

/*** code A1 : CMP_INDX ***/

int CMP_INDX()

{
A_INDX(CMP);
}

/*** code B1 : CMP_INDY ***/

int CMP_INDY()

{
A_INDY(CMP);
}

/*** code AD : CMP_ABS ***/

int CMP_ABS()

{
A_ABS(CMP);
}

/*** code BD : CMP_ABSX ***/

int CMP_ABSX()

{
A_ABSX(CMP);
}

/*** code B9 : CMP_ABSY ***/

int CMP_ABSY()

{
A_ABSY(CMP);
}

/***************************************************************/
/*******************    CPX     ********************************/
/***************************************************************/

/*** code E4 : CPX_ZRP ***/

int CPX_ZRP()

{
A_ZRP(CPX);
}

/*** code E0 : CPX_IMM ***/

int CPX_IMM()

{
A_IMM(CPX);
}


/*** code EC : CPX_ABS ***/

int CPX_ABS()

{
A_ABS(CPX);
}



/***************************************************************/
/*******************    CPY     ********************************/
/***************************************************************/

/*** code C4 : CPY_ZRP ***/

int CPY_ZRP()

{
A_ZRP(CPY);
}

/*** code C0 : CPY_IMM ***/

int CPY_IMM()

{
A_IMM(CPY);
}


/*** code CC : CPY_ABS ***/

int CPY_ABS()

{
A_ABS(CPY);
}



/***************************************************************/
/*******************    STA     ********************************/
/***************************************************************/

/*** code 85 : STA_ZRP ***/

int STA_ZRP()

{
Place(A,(unsigned short int)M[++PC]);
PC++;
return NEXT_INSTRUCTION;
}


/*** code 95 : STA_ZRPX ***/

int STA_ZRPX()

{
PC++;
Place(A,(unsigned short int)((M[PC]+X)&0xff));  /** Zero-Page indexe X **/
PC++;
return NEXT_INSTRUCTION;
}

/*** code 81 : STA_INDX ***/

int STA_INDX()

{
register unsigned char ADR;

ADR=M[++PC]+X;
Place(A,M[ADR]+(M[ADR+1]<<8));  /** Zero-Page indirect X **/
PC++;
return NEXT_INSTRUCTION;

/*PC++;
Place(A,(unsigned short int)(M[M[PC]]+(M[M[PC]+1]<<8)+X)); 
PC++;
return NEXT_INSTRUCTION;*/
}

/*** code 91 : STA_INDY ***/

int STA_INDY()

{
PC++;
Place(A,(unsigned short int)(M[M[PC]]+(M[M[PC]+1]<<8)+Y));  
PC++;
return NEXT_INSTRUCTION;
}

/*** code 8D : STA_ABS ***/

int STA_ABS()

{
PC++;
Place(A,(unsigned short int)(M[PC]+(M[PC+1]<<8)));  /** absolu **/
PC+=2;
return NEXT_INSTRUCTION;
}

/*** code 9D : STA_ABSX ***/

int STA_ABSX()

{
PC++;
Place(A,(unsigned short int)(M[PC]+(M[PC+1]<<8)+X));  /** absolu X **/
PC+=2;
return NEXT_INSTRUCTION;
}

/*** code 99 : STA_ABSY ***/

int STA_ABSY()

{
PC++;
Place(A,(unsigned short int)(M[PC]+(M[PC+1]<<8)+Y));  /** absolu Y **/
PC+=2;
return NEXT_INSTRUCTION;
}

/***************************************************************/
/**************** Codes de branchement *************************/
/***************************************************************/


/*** code 90 : BCC ***/

int BCC()

{
BxF(C_BIT);
}

/*** code B0 : BCS ***/

int BCS()

{
BxV(C_BIT);
}

/*** code F0 : BEQ ***/

int BEQ()

{
BxV(Z_BIT);
}

/*** code D0 : BNE ***/

int BNE()

{
BxF(Z_BIT);
}

/*** code 30 : BMI ***/

int BMI()

{
BxV(N_BIT);
}

/*** code 10 : BPL ***/

int BPL()

{
BxF(N_BIT);
}

/*** code 70 : BVS ***/

int BVS()

{
BxV(V_BIT);
}

/*** code 50 : BVC ***/

int BVC()

{
BxF(V_BIT);
}

/***************************************************************/
/*******************    STX     ********************************/
/***************************************************************/

/*** code 86 : STX_ZRP ***/

int STX_ZRP()

{
Place(X,(unsigned short int)M[++PC]);
PC++;
return NEXT_INSTRUCTION;
}

/*** code 86 : STX_ZRPY ***/

int STX_ZRPY()

{
Place(X,(unsigned short int)((M[++PC]+Y)&0xff));
PC++;
return NEXT_INSTRUCTION;
}


/*** code 8E : STX_ABS ***/

int STX_ABS()

{
PC++;
Place(X,(unsigned short int)(M[PC]+(M[PC+1]<<8)));  /** absolu **/
PC+=2;
return NEXT_INSTRUCTION;
}


/***************************************************************/
/*******************    STY     ********************************/
/***************************************************************/

/*** code 86 : STY_ZRP ***/

int STY_ZRP()

{
Place(Y,(unsigned short int)M[++PC]);
PC++;
return NEXT_INSTRUCTION;
}

/*** code 86 : STY_ZRPX ***/

int STY_ZRPX()

{
Place(Y,(unsigned short int)((M[++PC]+X)&0xff));
PC++;
return NEXT_INSTRUCTION;
}


/*** code 8E : STY_ABS ***/

int STY_ABS()

{
PC++;
Place(Y,(unsigned short int)(M[PC]+(M[PC+1]<<8)));  /** absolu **/
PC+=2;
return NEXT_INSTRUCTION;
}


/************* Positionnement des drapeaux ********************/

/*********************/
/*** code D8 : CLD ***/
/*********************/

int CLD()

{
CLx(D_BIT);
}


/*********************/
/*** code F8 : SED ***/
/*********************/

int SED()

{
SEx(D_BIT);
}

/*********************/
/*** code 58 : CLI ***/
/*********************/

int CLI()

{
CLx(I_BIT);
}


/*********************/
/*** code 78 : SEI ***/
/*********************/

int SEI()

{
SEx(I_BIT);
}

/*********************/
/*** code D8 : CLC ***/
/*********************/

int CLC()

{
CLx(C_BIT);
}


/*********************/
/*** code F8 : SEC ***/
/*********************/

int SEC()

{
SEx(C_BIT);
}

/*********************/
/*** code B8 : CLV ***/
/*********************/

int CLV()

{
CLx(V_BIT);
}

/*********************/
/*** code EA : NOP ***/
/*********************/

int NOP()

{
PC++;
NC=2;
return NEXT_INSTRUCTION;
}

/***************************************************************/
/****************** Transferts *********************************/
/***************************************************************/


/*** code AA : TAX ***/

int TAX()

{
Txx(A,X);
}

/*** code A8 : TAY ***/

int TAY()

{
Txx(A,Y);
}

/*** code BA : TSX ***/

int TSX()

{
Txx(S,X);
}

/*** code 8A : TXA ***/

int TXA()

{
Txx(X,A);
}

/*** code 98 : TYA ***/

int TYA()

{
Txx(Y,A);
}

/*** code 9A : TXS ***/

int TXS()

{
S=X;
PC++;
return NEXT_INSTRUCTION;
}

/*********************************************************************/
/********************* INX,INY,DEX,DEY *******************************/
/*********************************************************************/

/*** code E8 : INX ***/

int INX()

{
P&=(0xff^0x82);
if (!(++X)) P|=0x2;
if (X>127) P|=0x80;
PC++;
return NEXT_INSTRUCTION;
}

/*** code C8 : INY ***/

int INY()

{
P&=(0xff^0x82);
if (!(++Y)) P|=0x2;
if (Y>127) P|=0x80;
PC++;
return NEXT_INSTRUCTION;
}

/*** code CA : DEX ***/

int DEX()

{
P&=(0xff^0x82);
if (!(--X)) P|=0x2;
if (X>127) P|=0x80;
PC++;
return NEXT_INSTRUCTION;
}

/*** code 88 : DEY ***/

int DEY()

{
P&=(0xff^0x82);
if (!(--Y)) P|=0x2;
if (Y>127) P|=0x80;
PC++;
return NEXT_INSTRUCTION;
}

/*********************************************************************/
/********************* INC *******************************************/
/*********************************************************************/

/*** code E6 : INC_ZRP ***/

int INC_ZRP()

{
INC((unsigned short int)M[++PC]);
PC++;
return NEXT_INSTRUCTION;
}

/*** code F6 : INC_ZRPX ***/

int INC_ZRPX()

{
INC((unsigned short int)(M[++PC]+X)&0xff);
PC++;
return NEXT_INSTRUCTION;
}


/*** code EE : INC_ABS ***/

int INC_ABS()

{
PC++;
INC(M[PC]+(M[PC+1]<<8));  /** absolu **/
PC+=2;
return NEXT_INSTRUCTION;
}

/*** code FE : INC_ABSX ***/

int INC_ABSX()

{
PC++;
INC(M[PC]+(M[PC+1]<<8)+X);  /** absolu **/
PC+=2;
return NEXT_INSTRUCTION;
}


/*********************************************************************/
/********************* DEC *******************************************/
/*********************************************************************/

/*** code C6 : DEC_ZRP ***/

int DEC_ZRP()

{
DEC((unsigned short int)M[++PC]);
PC++;
return NEXT_INSTRUCTION;
}

/*** code D6 : DEC_ZRPX ***/

int DEC_ZRPX()

{
DEC((unsigned short int)(M[++PC]+X)&0xff);
PC++;
return NEXT_INSTRUCTION;
}


/*** code CE : DEC_ABS ***/

int DEC_ABS()

{
PC++;
DEC(M[PC]+(M[PC+1]<<8));  /** absolu **/
PC+=2;
return NEXT_INSTRUCTION;
}

/*** code DE : DEC_ABSX ***/

int DEC_ABSX()

{
PC++;
DEC(M[PC]+(M[PC+1]<<8)+X);  /** absolu **/
PC+=2;
return NEXT_INSTRUCTION;
}



/*********************************************************************/
/******************** Empilement/depilement **************************/
/*********************************************************************/

/*** code 48 : PHA ***/

int PHA()

{
Empiler(A);
NC=3;
PC++;
return NEXT_INSTRUCTION;
}

/*** code 08 : PHP ***/

int PHP()

{
Empiler(P);
NC=3;
PC++;
return NEXT_INSTRUCTION;
}

/*** code 68 : PLA ***/

int PLA()

{
P&=(0xff^0x82);
A=Depiler();
PC++;
NC=4;
if (!A) {P|=0x2;return NEXT_INSTRUCTION;}
if (A>127) P|=0x80;


return NEXT_INSTRUCTION;
}

/*** code 28 : PLP ***/

int PLP()

{
P=Depiler();
PC++;
NC=4;
return NEXT_INSTRUCTION;
}

/*********************************************************************/
/********************* ROL et ROR ************************************/
/*********************************************************************/

/*** code 2A : ROL implicite ***/

int ROL_IMP()

{
A=ROL(A);
PC++;
return NEXT_INSTRUCTION;
}

/*** code 26 : ROL Zero-Page ***/

int ROL_ZRP()

{
register unsigned short int ADR;

ADR=M[++PC];
Place(ROL(M[ADR]),ADR);
PC++;
return NEXT_INSTRUCTION;
}

/*** code 36 : ROL Zero-Page X ***/

int ROL_ZRPX()

{
register unsigned short int ADR;

ADR=(M[++PC]+X)&0xff;
Place(ROL(M[ADR]),ADR);
PC++;
return NEXT_INSTRUCTION;
}

/*** code 2E : ROL Absolu ***/

int ROL_ABS()

{
unsigned short int ADR;

PC++;
ADR=M[PC]+(M[PC+1]<<8);

Place(ROL(M[ADR]),ADR);
PC+=2;
return NEXT_INSTRUCTION;
}

/*** code 3E : ROL Absolu X ***/

int ROL_ABSX()

{
register unsigned short int ADR;

PC++;
ADR=M[PC]+(M[PC+1]<<8);

Place(ROL(M[ADR+X]),ADR+X);
PC+=2;
return NEXT_INSTRUCTION;
}



/*** code 2A : ROR implicite ***/

int ROR_IMP()

{
A=ROR(A);
PC++;
return NEXT_INSTRUCTION;
}


/*** code 26 : ROR Zero-Page ***/

int ROR_ZRP()

{
register unsigned short int m;


m=M[++PC];
Place(ROR(M[m]),m);
PC++;
return NEXT_INSTRUCTION;
}

/*** code 36 : ROR Zero-Page X ***/

int ROR_ZRPX()

{
register unsigned short int ADR;


ADR=(M[++PC]+X)&0xff;
Place(ROR(M[ADR]),ADR);
PC++;
return NEXT_INSTRUCTION;
}

/*** code 2E : ROR Absolu ***/

int ROR_ABS()

{
register unsigned short int ADR;

PC++;
ADR=M[PC]+(M[PC+1]<<8);

Place(ROR(M[ADR]),ADR);
PC+=2;
return NEXT_INSTRUCTION;
}

/*** code 3E : ROR Absolu X ***/

int ROR_ABSX()

{
register unsigned short int ADR;

PC++;
ADR=M[PC]+(M[PC+1]<<8)+X;

Place(ROR(M[ADR]),ADR);
PC+=2;
return NEXT_INSTRUCTION;
}


/*********************************************************************/
/********************* ASL et LSR ************************************/
/*********************************************************************/

/*** code 0A : ASL implicite ***/

int ASL_IMP()

{
A=ASL(A);
PC++;
return NEXT_INSTRUCTION;
}

/*** code 06 : ASL Zero-Page ***/

int ASL_ZRP()

{
PC++;
Place(ASL(M[M[PC]]),(unsigned short int)M[PC]);
PC++;
return NEXT_INSTRUCTION;
}

/*** code 16 : ASL Zero-Page X ***/

int ASL_ZRPX()

{
PC++;
Place(ASL(M[(M[PC]+X)&0xff]),(unsigned short int)((M[PC]+X)&0xff));
PC++;
return NEXT_INSTRUCTION;
}

/*** code 0E : ASL Absolu ***/

int ASL_ABS()

{
unsigned short int ADR;

PC++;
ADR=M[PC]+(M[PC+1]<<8);

Place(ASL(M[ADR]),ADR);
PC+=2;
return NEXT_INSTRUCTION;
}

/*** code 1E : ASL Absolu X ***/

int ASL_ABSX()

{
register unsigned short int ADR;

PC++;
ADR=M[PC]+(M[PC+1]<<8);

Place(ASL(M[ADR+X]),ADR+X);
PC+=2;
return NEXT_INSTRUCTION;
}



/*** code 4A : LSR implicite ***/

int LSR_IMP()

{
A=LSR(A);
PC++;
return NEXT_INSTRUCTION;
}



/*** code 46 : LSR Zero-Page ***/

int LSR_ZRP()

{
PC++;
Place(LSR(M[M[PC]]),M[PC]);
PC++;
return NEXT_INSTRUCTION;
}

/*** code 56 : LSR Zero-Page X ***/

int LSR_ZRPX()

{
PC++;
Place(LSR(M[(M[PC]+X)&0xff]),(unsigned short int)((M[PC]+X)&0xff));
PC++;
return NEXT_INSTRUCTION;
}

/*** code 4E : LSR Absolu ***/

int LSR_ABS()

{
unsigned short int ADR;

PC++;
ADR=M[PC]+(M[PC+1]<<8);

Place(LSR(M[ADR]),ADR);
PC+=2;
return NEXT_INSTRUCTION;
}

/*** code 5E : LSR Absolu X ***/

int LSR_ABSX()

{
unsigned short int ADR;

PC++;
ADR=M[PC]+(M[PC+1]<<8);

Place(LSR(M[ADR+X]),ADR+X);
PC+=2;
return NEXT_INSTRUCTION;
}



/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

/**** initialise les fonctions ****/

void InitFonctions()

{
int i;

for (i=0;i<256;i++) PFONCT[i]=IllegalInstruction;

PFONCT[0x00]=BRK;
PFONCT[0x20]=JSR;
PFONCT[0x4C]=JMP_IMM;
PFONCT[0x6C]=JMP_IND;

PFONCT[0x60]=RTS;     /* Return From Subroutine */
PFONCT[0x40]=RTI;     /* Return From Interrupt */

PFONCT[0x65]=ADC_ZRP;
PFONCT[0x69]=ADC_IMM;
PFONCT[0x75]=ADC_ZRPX;
PFONCT[0x61]=ADC_INDX;
PFONCT[0x71]=ADC_INDY;
PFONCT[0x6D]=ADC_ABS;
PFONCT[0x7D]=ADC_ABSX;
PFONCT[0x79]=ADC_ABSY;

PFONCT[0xE5]=SBC_ZRP;
PFONCT[0xE9]=SBC_IMM;
PFONCT[0xF5]=SBC_ZRPX;
PFONCT[0xE1]=SBC_INDX;
PFONCT[0xF1]=SBC_INDY;
PFONCT[0xED]=SBC_ABS;
PFONCT[0xFD]=SBC_ABSX;
PFONCT[0xF9]=SBC_ABSY;

PFONCT[0x24]=BIT_ZRP;
PFONCT[0x2C]=BIT_ABS;

PFONCT[0x81]=STA_INDX;     /* STA Indirect X */
PFONCT[0x8D]=STA_ABS;      /* STA Absolu */
PFONCT[0x9D]=STA_ABSX;     /* STA Absolu X */
PFONCT[0x99]=STA_ABSY;     /* STA Absolu Y */
PFONCT[0x91]=STA_INDY;     /* STA Indirect Y */
PFONCT[0x85]=STA_ZRP;      /* STA Zero-Page */
PFONCT[0x95]=STA_ZRPX;     /* STA Zero-Page X */

PFONCT[0x86]=STX_ZRP;      /* STX Zero-Page */
PFONCT[0x96]=STX_ZRPY;     /* STX Zero-Page Y */
PFONCT[0x8E]=STX_ABS;      /* STX Absolu */

PFONCT[0x84]=STY_ZRP;      /* STY Zero-Page */
PFONCT[0x94]=STY_ZRPX;     /* STY Zero-Page X */
PFONCT[0x8C]=STY_ABS;      /* STY Absolu */

PFONCT[0xA0]=LDY_IMM;   /* LDY Immediat */
PFONCT[0xA4]=LDY_ZRP;   /* LDY Zero-Page */
PFONCT[0xB4]=LDY_ZRPX;  /* LDY Zero-Page X */
PFONCT[0xAC]=LDY_ABS;   /* LDY absolu */
PFONCT[0xBC]=LDY_ABSX;  /* LDY absolu X */

PFONCT[0xA2]=LDX_IMM;   /* LDX Immediat */
PFONCT[0xA6]=LDX_ZRP;   /* LDX Zero-Page */
PFONCT[0xB6]=LDX_ZRPY;  /* LDX Zero-Page Y */
PFONCT[0xAE]=LDX_ABS;   /* LDX absolu */
PFONCT[0xBE]=LDX_ABSY;  /* LDX absolu Y */

PFONCT[0xA9]=LDA_IMM;  /* LDA Immediat */
PFONCT[0xA5]=LDA_ZRP;  /* LDA Zero-Page */
PFONCT[0xB5]=LDA_ZRPX; /* LDA Zero-Page X */
PFONCT[0xA1]=LDA_INDX; /* LDA Indirect X */
PFONCT[0xB1]=LDA_INDY; /* LDA Indirect Y */
PFONCT[0xAD]=LDA_ABS;  /* LDA absolu */
PFONCT[0xBD]=LDA_ABSX; /* LDA absolu X */
PFONCT[0xB9]=LDA_ABSY; /* LDA absolu Y */

PFONCT[0x29]=AND_IMM;  /* AND Immediat */
PFONCT[0x25]=AND_ZRP;  /* AND Zero-Page */
PFONCT[0x35]=AND_ZRPX; /* AND Zero-Page X */
PFONCT[0x21]=AND_INDX; /* AND Indirect X */
PFONCT[0x31]=AND_INDY; /* AND Indirect Y */
PFONCT[0x2D]=AND_ABS;  /* AND absolu */
PFONCT[0x3D]=AND_ABSX; /* AND absolu X */
PFONCT[0x39]=AND_ABSY; /* AND absolu Y */

PFONCT[0x49]=EOR_IMM;  /* EOR Immediat */
PFONCT[0x45]=EOR_ZRP;  /* EOR Zero-Page */
PFONCT[0x55]=EOR_ZRPX; /* EOR Zero-Page X */
PFONCT[0x41]=EOR_INDX; /* EOR Indirect X */
PFONCT[0x51]=EOR_INDY; /* EOR Indirect Y */
PFONCT[0x4D]=EOR_ABS;  /* EOR absolu */
PFONCT[0x5D]=EOR_ABSX; /* EOR absolu X */
PFONCT[0x59]=EOR_ABSY; /* EOR absolu Y */

PFONCT[0x09]=ORA_IMM;  /* ORA Immediat */
PFONCT[0x05]=ORA_ZRP;  /* ORA Zero-Page */
PFONCT[0x15]=ORA_ZRPX; /* ORA Zero-Page X */
PFONCT[0x01]=ORA_INDX; /* ORA Indirect X */
PFONCT[0x11]=ORA_INDY; /* ORA Indirect Y */
PFONCT[0x0D]=ORA_ABS;  /* ORA absolu */
PFONCT[0x1D]=ORA_ABSX; /* ORA absolu X */
PFONCT[0x19]=ORA_ABSY; /* ORA absolu Y */


PFONCT[0xC5]=CMP_ZRP;  /* CMP Zero-Page */
PFONCT[0xC9]=CMP_IMM;  /* CMP Immediat */
PFONCT[0xD1]=CMP_INDY; /* CMP Indirect Y */
PFONCT[0xC1]=CMP_INDX; /* CMP Indirect X */
PFONCT[0xD5]=CMP_ZRPX; /* CMP Zero-Page X */
PFONCT[0xCD]=CMP_ABS;  /* CMP absolu */
PFONCT[0xDD]=CMP_ABSX; /* CMP absolu X */
PFONCT[0xD9]=CMP_ABSY; /* CMP absolu Y */

PFONCT[0xE0]=CPX_IMM; /* CPX immediat */
PFONCT[0xE4]=CPX_ZRP; /* CPX immediat */
PFONCT[0xEC]=CPX_ABS; /* CPX absolu */

PFONCT[0xC0]=CPY_IMM; /* CPY immediat */
PFONCT[0xC4]=CPY_ZRP; /* CPY immediat */
PFONCT[0xCC]=CPY_ABS; /* CPY absolu */

PFONCT[0x90]=BCC; /* Branch on carry clear */
PFONCT[0xB0]=BCS; /* Branch on carry set   */

PFONCT[0xF0]=BEQ; /* Branch on 0 */
PFONCT[0xD0]=BNE; /* Branch on non 0   */

PFONCT[0x30]=BMI; /* Branch on - */
PFONCT[0x10]=BPL; /* Branch on + */

PFONCT[0x50]=BVC; /* Branch on V=0 */
PFONCT[0x70]=BVS; /* Branch on V=1 */


PFONCT[0xD8]=CLD; /* Clear Decimal */
PFONCT[0xF8]=SED; /* Set Decimal */

PFONCT[0x18]=CLC; /* Clear Carry */
PFONCT[0x38]=SEC; /* Set Carry */

PFONCT[0x58]=CLI; /* Clear Insterrupt Mask */
PFONCT[0x78]=SEI; /* Set Insterrupt Mask */

PFONCT[0xB8]=CLV; /* Clear Overflow Flag */

PFONCT[0xEA]=NOP; /* No Operation */

PFONCT[0x8A]=TXA; /* X->A */
PFONCT[0x9A]=TXS; /* X->S */
PFONCT[0x98]=TYA; /* Y->A */

PFONCT[0xBA]=TSX; /* S->X */
PFONCT[0xA8]=TAY; /* A->Y */
PFONCT[0xAA]=TAX; /* A->X */

PFONCT[0xE8]=INX; /* X++ */
PFONCT[0xC8]=INY; /* Y++ */
PFONCT[0xCA]=DEX; /* X-- */
PFONCT[0x88]=DEY; /* Y-- */

PFONCT[0x48]=PHA; /* Push A onto Stack */
PFONCT[0x08]=PHP; /* Push P onto Stack */
PFONCT[0x68]=PLA; /* Pull A from Stack */
PFONCT[0x28]=PLP; /* Pull P from Stack */

PFONCT[0x2A]=ROL_IMP;  /* ROL Implicite */
PFONCT[0x26]=ROL_ZRP;  /* ROL Zero Page */
PFONCT[0x36]=ROL_ZRPX; /* ROL Zero Page X */
PFONCT[0x2E]=ROL_ABS;  /* ROL Absolu */
PFONCT[0x3E]=ROL_ABSX; /* ROL Absolu X */

PFONCT[0x6A]=ROR_IMP;  /* ROR Implicite */
PFONCT[0x66]=ROR_ZRP;  /* ROR Zero Page */
PFONCT[0x76]=ROR_ZRPX; /* ROR Zero Page X */
PFONCT[0x6E]=ROR_ABS;  /* ROR Absolu */
PFONCT[0x7E]=ROR_ABSX; /* ROR Absolu X */

PFONCT[0x0A]=ASL_IMP;  /* ASL Implicite */
PFONCT[0x06]=ASL_ZRP;  /* ASL Zero Page */
PFONCT[0x16]=ASL_ZRPX; /* ASL Zero Page X */
PFONCT[0x0E]=ASL_ABS;  /* ASL Absolu */
PFONCT[0x1E]=ASL_ABSX; /* ASL Absolu X */

PFONCT[0x4A]=LSR_IMP;  /* LSR Implicite */
PFONCT[0x46]=LSR_ZRP;  /* LSR Zero Page */
PFONCT[0x56]=LSR_ZRPX; /* LSR Zero Page X */
PFONCT[0x4E]=LSR_ABS;  /* LSR Absolu */
PFONCT[0x5E]=LSR_ABSX; /* LSR Absolu X */

PFONCT[0xC6]=DEC_ZRP;   /* DEC Zero-Page */
PFONCT[0xD6]=DEC_ZRPX;  /* DEC Zero-Page X */
PFONCT[0xCE]=DEC_ABS;   /* DEC Absolu */
PFONCT[0xDE]=DEC_ABSX;  /* DEC Absolu X */

PFONCT[0xE6]=INC_ZRP;   /* INC Zero-Page */
PFONCT[0xF6]=INC_ZRPX;  /* INC Zero-Page X */
PFONCT[0xEE]=INC_ABS;   /* INC Absolu */
PFONCT[0xFE]=INC_ABSX;  /* INC Absolu X */

PFONCT[0x07]=TrapOPCode;  /* 07 a la place de certaines adresses ROM */
PFONCT[0x17]=TrapOPCode;  /* 17 a la place de certaines adresses ROM (pour breakpoints) */
PFONCT[0x27]=TrapOPCode;  /* 27 a la place de certaines adresses ROM */
PFONCT[0x37]=TrapOPCode;  /* 37 TEXT/HIRES */
PFONCT[0xFF]=TrapOPCode;  /* FF a la place de certaines adresses ROM */
}

