/***************************************************************************/
/**                                                                       **/
/**                 The ORIC 48K emulator.UNIX version                    **/
/**                                                                       **/
/**                Written by JF FABRE (fabre@supaero.fr)                 **/
/**                                                                       **/
/**                This program may be freely distibuted                  **/
/**                                                                       **/
/**             as long as this header is included unmodified             **/
/**                                                                       **/
/***************************************************************************/

/** Constants and structures definition **/

/** Modes video **/

#define TEXT 0
#define HIRES 1
#define HTHH 2

/** Debut de la zone de pile **/

#define STACK_BEGIN 0x100

/** Status processeur **/

#define ILLEGAL_INSTRUCTION 0
#define BREAKPOINT 1
#define USER_BREAK 2
#define NEXT_INSTRUCTION 3
#define MONITOR_COMMAND 4
#define EXIT_MONITOR_MODE 5
#define EXIT_EMULATOR 6
#define STEP_COMMAND 7
#define RESET_EMULATOR 8
#define STEP_OVER_COMMAND 9

#define USER_RESET1 10
#define USER_RESET2 11
#define CHANGE_TAPE 13

#define PATCH_OUTBYT 14
#define PATCH_RDBYTE 15
#define PATCH_SYNC 16
#define PATCH_OUTLED 17

#define PATCH_PRINT 18

/** Patches ROM **/

#define J_OUTBYT 0xE65E
#define J_RDBYTE 0xE6C9
#define J_SYNC   0xE735

#define J_OUTCHR 0xF5C1

#define J_RETURN_INIT 0xF8AF
#define J_RETURN_HIRES 0xF966

#define J_PING 0xFA9F
#define J_ZAP 0xFAE1
#define J_EXPLODE 0xFACB
#define J_SHOOT 0xFAB5
#define J_CONTBP 0xFB2A
#define J_KBBEEP 0xFB14
#define J_OLD_KBBEEP 0xFB10
#define J_SOUND 0xFB40
#define J_MUSIC 0xFC18
#define J_PLAY 0xFBD0


#define NO_ATT 0
#define PAPER_ATT 1
#define STD_ATT 2
#define ALT_ATT 3
#define FLASH_STD_ATT 4
#define FLASH_ALT_ATT 5
#define INK_ATT 6
#define STD_DBL_HEIGHT_ATT 7
#define ALT_DBL_HEIGHT_ATT 8
#define FLASH_STD_DBL_HEIGHT_ATT 9
#define FLASH_ALT_DBL_HEIGHT_ATT 10
#define GFX1_ATT 11
#define GFX2_ATT 12
#define TEXT1_ATT 13
#define TEXT2_ATT 14


#define N_MAX_BREAKPOINTS 10
#define OPL 16
#define LG_CMD 80
#define POINT_BUF_SIZE 5000

#include <math.h>
#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>  
#include <X11/keysym.h>

typedef struct
 {
 unsigned char ink,paper,flash,charmode,dblh;
 }
 contx;

typedef struct
 {
 Display *dpy;
 GC gc;
 /*Pixmap pixbuff;*/
 Window fen;
 int ecr;
 unsigned int coul[8];
 Colormap cmap;
 char schem_clav[32]; 
 }
 ecran;
 
 
typedef struct
 {
 unsigned short int BreakPointList[N_MAX_BREAKPOINTS]; 
 unsigned char BreakPointMem[N_MAX_BREAKPOINTS]; 
 int nb_break;
 
 unsigned short int MemWatchWMList[N_MAX_BREAKPOINTS]; 
 unsigned char MemWatchWMMem[N_MAX_BREAKPOINTS]; 
 unsigned char MWLastValue[N_MAX_BREAKPOINTS]; 
 int nb_mwWM;
 
 unsigned short int MemWatchRMList[N_MAX_BREAKPOINTS]; 
 unsigned char MemWatchRMMem[N_MAX_BREAKPOINTS]; 
 int nb_mwRM;
 
 int Mon_Status;
 char LastCmd[LG_CMD];
 char TapeLoadName[40],TapeSaveName[40];
 char TapeDir[100];
 char DumpDir[100];
 char KeybFileName[40];
 FILE *TapeLoad,*TapeSave;
 
 int KeyMat[8][8];
 char corresp[256];
 
 int dispmode;
 int sound,autorefresh,affiche;
 int BreakReq,interrupt,illeg_inst;
 FILE *printer_file;
 int hom_x,hom_y;
 }
 Monitor_struct;
 
 typedef struct
 {
 char no_reg,no_reg_req,val_req;
 char CA2,CB2,CA1,CB1;
 }
 CHIP8912Status_struct;
 
 typedef struct
 
 {
 unsigned char *buff;
 unsigned long int taille;
 }
 son;
 
 typedef struct
 {
 son son_beep,son_contbeep,son_zap,son_shoot,son_explode,son_ping;
 }
 SND;
 
 typedef struct
 {
 int id;
 int fd_pere[2],fd_fils[2];
 }
 comm_struct;
 
typedef int (*T_PFONCT)();
