/***************************************************************************/
/**                                                                       **/
/**                 The ORIC 48K emulator.UNIX version                    **/
/**                                                                       **/
/**                Written by JF FABRE (fabre@supaero.fr)                 **/
/**                                                                       **/
/**                This program may be freely distibuted                  **/
/**                                                                       **/
/**             as long as this header is included unmodified             **/
/**                                                                       **/
/***************************************************************************/

/* In-line command interpreter */

#include "protos.h"

extern unsigned short int PC;
extern unsigned char S,X,A,Y,P;
extern unsigned char *M;
extern ecran disp;

extern Monitor_struct mon;
extern CHIP8912Status_struct C8912S;

#define MAX_ARGS 6

typedef struct
 {
 char ch[LG_CMD];
 }
 str;

/**** lire le nb d'arguments ****/

int ReadArgs(ch,chi)

char *ch;
str *chi;

{
char *ch2;
int i,j;

ch2=ch;

while(((*ch2)!='\0')&&((*ch2)==' ')) ch2++; /* vire les espaces */

if (!strlen(ch2)) strcpy(ch2,mon.LastCmd); else strcpy(mon.LastCmd,ch2);
 
for (i=0;i<MAX_ARGS;i++)
  {
  j=0;
  while(((*ch2)!='\0')&&((*ch2)!=' ')) {chi[i].ch[j]=*ch2; ch2++;j++;}
  chi[i].ch[j]='\0';
  if ((*ch2)=='\0') return i;
  while(((*ch2)!='\0')&&((*ch2)==' ')) ch2++; /* vire les espaces */
  }
  
return i;  
}

/******** Read Hex bytes ***************/

int ReadHexBytes(ch,ht)

char *ch,*ht;

{
int i,nbbytes=0;

while(((*ch)!='\0')&&((*ch)==' ')) ch++; /* vire les espaces */
if (ch[0]=='\0') return 0;

if (ch[0]==34)
   {
   ch++; /* skip du " */ 
   /** on lit de l'ASCII **/

   while(((*ch)!='\0')&&((*ch)!=34)) {*(ht++)=*(ch++);nbbytes++;}
   }

else

   for (;;)
   {
   /** on lit de l'HEXA **/

   ht[nbbytes++]=(unsigned char)ValeurHexa(ch);

   while(((*ch)!='\0')&&((*ch)!=' ')) ch++; /* recherche un espace */

   while(((*ch)!='\0')&&((*ch)==' ')) ch++; /* vire les espaces */
   if (ch[0]=='\0') break;
   }

return nbbytes;
}

/******** instuctions moniteur *********/

int MonitorInterpreter(ch)

char *ch;
{
char ch2[LG_CMD],ch3[LG_CMD],c;
str chi[MAX_ARGS];
unsigned int i,j,nargs,nbbytes,nb;
unsigned short int opcode,opcode2,opcode3;

FILE *f;


  if (ch[0]=='!') 
  {
  system(ch+1);
  return MONITOR_COMMAND;
  }
  
nargs=ReadArgs(ch,chi);

 if (!strncmp(chi[0].ch,"mem",3))
 {
 
 switch(nargs)
   {
   case 0: opcode=PC;opcode2=opcode+16;break;
   case 1: opcode=ValeurHexa(chi[1].ch);opcode2=opcode+16;break;
   default: opcode=ValeurHexa(chi[1].ch);opcode2=ValeurHexa(chi[2].ch);
   }
 
 DumpMemory(opcode,opcode2);
 return MONITOR_COMMAND;
 }

 if ((!strncmp(chi[0].ch,"exi",3))||(!strncmp(chi[0].ch,"qui",3)))
 {
 return EXIT_EMULATOR;
 }

 if (!strncmp(chi[0].ch,"reset",5))
 {
 return RESET_EMULATOR;
 }

 if (!strncmp(chi[0].ch,"flu",3))
 {
 RefreshDisplay();
 return MONITOR_COMMAND;
 }

 if (!strncmp(chi[0].ch,"rew",3))
 {
 if (!RewindTapeLoad())
   printf("Sorry, no tape specified !\n");
 else
   printf("TapeLoad rewinded.\n");
   
 return MONITOR_COMMAND;
 }
 

 if (!strncmp(chi[0].ch,"re",2))
 {
 switch(nargs)
   {
   case 0:
     PrintRegisters();
     return MONITOR_COMMAND;
 
   case 1:
     printf("** Syntax error\n"); 
     return MONITOR_COMMAND;
     
   default:
     switch(chi[1].ch[0])
  {
  case 'a' :
  case 'A' : A=ValeurHexa(chi[2].ch);
             break;
  case 'x' :
  case 'X' : X=ValeurHexa(chi[2].ch);
             break;
  case 'y' :
  case 'Y' : Y=ValeurHexa(chi[2].ch);
             break;
  case 's' :
  case 'S' : S=ValeurHexa(chi[2].ch);
             break;
  case 'p' :
  case 'P' : if ((chi[1].ch[1]=='c')||(chi[1].ch[1]=='C'))
               PC=ValeurHexa(chi[2].ch);
              else
               P=ValeurHexa(chi[2].ch);
             break;

  default : printf("Not a valid register.\n");
  
  }           
  }
  
  PrintRegisters();
  return MONITOR_COMMAND;
             
 
 }

 if (!strncmp(chi[0].ch,"ass",3))
 {
 printf("Assembler mode not implemented\n");
 return MONITOR_COMMAND;

 }

 if (!strncmp(chi[0].ch,"hel",3))
 {
 if (!(f=fopen("Monitor.help","rt"))) printf("Sorry, no help available !\n");
  else
  {
  while (!feof(f))   
   {
   c=fgetc(f);
   if (c>0) putchar(c);
   }
   
  fclose(f); 
  }
 return MONITOR_COMMAND;
 }

 if (!strncmp(chi[0].ch,"st",2))
 {
 return STEP_COMMAND;
 }

 if (!strncmp(chi[0].ch,"tr",2))
 {
 return STEP_OVER_COMMAND;
 }

 if (!strncmp(chi[0].ch,"med",3))
 {
 switch(nargs)
   {
   case 0:  opcode=PC;break;
   default: opcode=ValeurHexa(chi[1].ch);
   }
 
 if (opcode>0xFFFF) 
   {
   printf("Memory adress must be between 0x0000 and 0xFFFF !\n");
   return MONITOR_COMMAND;
   }
   else
   {
   MemoryEditMode(opcode);
   return MONITOR_COMMAND;
   }
 

 }

 if (!strncmp(chi[0].ch,"des",3))
 {
 switch(nargs)
   {
   case 0:  opcode=PC;break;
   default: opcode=ValeurHexa(chi[1].ch);
   }
 
 if (opcode>0xFFFF) 
   {
   printf("Disassemble adress must be between 0x0000 and 0xFFFF !\n");
   return MONITOR_COMMAND;
   }
   else
   {
   DisassembleMode(opcode);
   return MONITOR_COMMAND;
   }
 

 }

 if (!strncmp(chi[0].ch,"ldm",3))
 {
 switch(nargs)
   {
   case 0: printf("Filename expected.\n");return MONITOR_COMMAND;
   case 1: printf("Start address expected.\n");return MONITOR_COMMAND;
   }
   
   opcode=ValeurHexa(chi[2].ch);
   
   sprintf(ch2,"%s/%s",mon.DumpDir,chi[1].ch);
     
 if (!(f=fopen(ch2,"rb")))  
    {
    printf("%s : File open error\n",chi[1].ch);
    return MONITOR_COMMAND;
    }

  c=fgetc(f);
     
  for (opcode2=opcode;opcode2<=0xFFFF;opcode2++)
   {
   if (feof(f)) break;
   
   if (opcode2<0xC000) Place(c,opcode2);
                  else M[opcode2]=c;
   c=fgetc(f);
   }
         
  fclose(f);

  
  printf("File Read from ");PrintAddress(opcode);
  printf(" to ");PrintAddress(opcode2);putchar('\n');

  if (opcode2>=0xC000)
    printf("Warning : ROM modified\n");
  
  return MONITOR_COMMAND;
 }
 
 if (!strncmp(chi[0].ch,"svm",3))
 {
 switch(nargs)
   {
   case 0: printf("Filename expected.\n");return MONITOR_COMMAND;
   case 1: printf("Start address expected.\n");return MONITOR_COMMAND;
   case 2: printf("Ending address expected.\n");return MONITOR_COMMAND;
   }
   
   opcode=ValeurHexa(chi[2].ch);
   opcode2=ValeurHexa(chi[3].ch);

 sprintf(ch2,"%s/%s",mon.DumpDir,chi[1].ch);
     
 if ((f=fopen(ch2,"rb")))  
    {
    fclose(f);
    printf("%s : File already exists. ",chi[1].ch);
    if (!Ask("Replace"))
       {
       printf("** Canceled\n");
       return MONITOR_COMMAND;
       }
    }
     
 if (!(f=fopen(ch2,"wb")))  
    {
    printf("** %s : File creation error\n",chi[1].ch);
    return MONITOR_COMMAND;
    }

     
  for (i=opcode;i<opcode2;i++)
   fputc(M[i],f);
   
  fclose(f);

  
  printf("File Written ");PrintAddress(opcode);
  printf(" to ");PrintAddress(opcode2);printf(".\nLength : ");
  PrintAddress(opcode2-opcode);putchar('\n');
  
  return MONITOR_COMMAND;
 }
 
  if (!strncmp(chi[0].ch,"lsn",3))
 {

 if (!nargs) {printf("Filename expected.\n");return MONITOR_COMMAND;}

 sprintf(ch2,"%s/%s",mon.DumpDir,chi[1].ch);
     
 if (!(f=fopen(ch2,"rb")))  
    {
    printf("%s : File open error\n",chi[1].ch);
    return MONITOR_COMMAND;
    }

   
  fread(M, 1, 0xC000, f);
  
  A=fgetc(f);
  X=fgetc(f);
  Y=fgetc(f);
  P=fgetc(f);
  S=fgetc(f);
  PC=(unsigned short int)fgetc(f);
  PC+=((unsigned short int)fgetc(f))<<8;
  mon.dispmode=fgetc(f);
     
  fclose(f);
  
  printf("Snapshot Read.\n");
  RefreshDisplay();
    
  return MONITOR_COMMAND;
 }
 

  if (!strncmp(chi[0].ch,"ssn",3))
 {
 if (!nargs) {printf("Filename expected.\n");return MONITOR_COMMAND;}

 sprintf(ch2,"%s/%s",mon.DumpDir,chi[1].ch);
     
     
 if (!(f=fopen(ch2,"wb")))  
    {
    printf("%s : File creation error\n",chi[1].ch);
    return MONITOR_COMMAND;
    }

  fwrite(M, 1, 0xC000, f);

  fputc(A,f);
  fputc(X,f);
  fputc(Y,f);
  fputc(P,f);
  fputc(S,f);
  fputc((unsigned char)(PC&0xFF),f);
  fputc((unsigned char)(PC>>8),f);
  fputc((unsigned char)mon.dispmode,f);
      
  
  fclose(f);
  
  printf("Snapshot written.\n");
    
  return MONITOR_COMMAND;
 }
 

 if (!strncmp(chi[0].ch,"go",2))
 {


 if (!nargs) opcode=PC; else opcode=ValeurHexa(chi[1].ch);
 
 if (opcode>0xFFFF) 
   {
   printf("Jump adress must be between 0x0000 and 0xFFFF !\n");
   return MONITOR_COMMAND;
   }
   else
   {
   PC=opcode;
   return EXIT_MONITOR_MODE;
   }
 }

 
 if (!strncmp("bs",ch,2))
 {

 if (!nargs) opcode=PC; else opcode=ValeurHexa(chi[1].ch);
 
 if (opcode>0xFFFF) 
   {
   printf("Breakpoint adress must be between 0x0000 and 0xFFFF !\n");
   }
   else
   {
   if (AddBreakPoint(opcode))  {printf("Breakpoint Set at ");PrintAddress(opcode);putchar('\n');}
                          else printf("Too many breakpoints !\n");
   }

return MONITOR_COMMAND;
 }

 if (!strncmp("mws",ch,2))
 {
 if (!nargs) opcode=PC; else opcode=ValeurHexa(chi[1].ch);

 
 if (opcode>0xFFFF) 
   {
   printf("Memwatchpoint adress must be between 0x0000 and 0xFFFF !\n");
   }
   else
   {
   if (AddMemWatchWM(opcode))  {printf("Memwatchpoint (write) Set at ");PrintAddress(opcode);putchar('\n');}
                          else printf("Too many memwatchpoints (write) !\n");
   }

return MONITOR_COMMAND;
 }

 if (!strncmp("mwd",ch,2))
 {
 if (!nargs) opcode=PC; else opcode=ValeurHexa(chi[1].ch);
 
 if (opcode>0xFFFF)
   {
   printf("Memwatchpoint adress must be between 0x0000 and 0xFFFF !\n");
   }
   else
   {
   if (RemoveMemWatchWM(opcode)) printf("Memwatchpoint Removed\n");
                            else printf("Memwatchpoint non-existent\n");
   }
 return MONITOR_COMMAND;
 }

 if (!strncmp("mrd",ch,2))
 {
 if (!nargs) opcode=PC; else opcode=ValeurHexa(chi[1].ch);
 
 if (opcode>0xFFFF)
   {
   printf("Memwatchpoint adress must be between 0x0000 and 0xFFFF !\n");
   }
   else
   {
   if (RemoveMemWatchRM(opcode)) printf("Memwatchpoint Removed\n");
                            else printf("Memwatchpoint non-existent\n");
   }
 return MONITOR_COMMAND;
 }
 

 if (!strncmp("mrs",ch,2))
 {
 if (!nargs) opcode=PC; else opcode=ValeurHexa(chi[1].ch);

 if (opcode>0xFFFF) 
   {
   printf("Memwatchpoint adress must be between 0x0000 and 0xFFFF !\n");
   }
   else
   {
   if (AddMemWatchRM(opcode))  {printf("Memwatchpoint (read) Set at ");PrintAddress(opcode);putchar('\n');}
                          else printf("Too many memwatchpoints (read) !\n");
   }

return MONITOR_COMMAND;
 }

 if (!strncmp("lb",ch,2))
 {
 ListBreakpoints();
 return MONITOR_COMMAND;
 }

 if (!strncmp("lmw",ch,2))
 {
 ListBreakpoints();
 return MONITOR_COMMAND;
 }

 if (!strncmp("lmr",ch,2))
 {
 ListBreakpoints();
 return MONITOR_COMMAND;
 }

 if (!strncmp(chi[0].ch,"bda",3))
 {
 DeleteBreakpoints();
 return MONITOR_COMMAND;
 }
 
 if (!strncmp(chi[0].ch,"mwda",4))
 {
 DeleteMemWatchesWM();
 return MONITOR_COMMAND;
 }
 
 if (!strncmp(chi[0].ch,"mrda",4))
 {
 DeleteMemWatchesRM();
 return MONITOR_COMMAND;
 }
 
 if (!strncmp("bd",ch,2))
 {
 if (!nargs) opcode=PC; else opcode=ValeurHexa(chi[1].ch);

 
 if (opcode>0xFFFF)
   {
   printf("Breakpoint adress must be between 0x0000 and 0xFFFF !\n");
   }
   else
   {
   if (RemoveBreakPoint(opcode)) printf("Breakpoint Removed\n");
                            else printf("Breakpoint non-existent\n");
   }
 return MONITOR_COMMAND;
 }
 
 if (!strncmp(chi[0].ch,"cht",3))
 { 
 if (nargs) ChangeTapeLoad(chi[1].ch); else ChangeTapeLoad("");
 return MONITOR_COMMAND;
 }
 
 if (!strncmp(chi[0].ch,"ost",3))
 { 
 if (nargs) ChangeTapeSave(chi[1].ch); else ChangeTapeSave("");
 return MONITOR_COMMAND;
 }
 
 if (!strncmp(chi[0].ch,"?",1))
 { 
 if (nargs) PrintExpression(chi[1].ch); else printf("** Syntax : eval [expression]...\n");  
 return MONITOR_COMMAND;
 }

 if (!strncmp(chi[0].ch,"fa",2))
 {
  switch(nargs)
    {
    case 0:printf("Usage : fa [adress] ([startaddr])\n");return MONITOR_COMMAND;
    case 1:opcode=ValeurHexa(chi[1].ch);opcode2=0;break;
    default:opcode2=ValeurHexa(chi[2].ch);opcode=ValeurHexa(chi[1].ch);
    }
    
  for (i=opcode2;i<=0xFFFF;i++)
   {
   Disass(i,ch2,M);
   strcpy(ch3,ch2);
   
   for (j=0;j<strlen(ch3);j++)
     if (ch3[j]=='$') break;
   
   if (j!=strlen(ch3)) 
      {
      ch3[j-1]='0';
      ch3[j]='x';
      sscanf(ch3+j-1,"%x",&nb);opcode3=nb;
      if (opcode3==opcode) 
          {
          PrintAddress((unsigned short)i);
          printf(" : %s\n",ch2);
          }
      }
      
   if (mon.BreakReq) {mon.BreakReq=0;break;}
   }
   
  printf("Searched from ");PrintAddress(opcode2);
  printf(" to ");PrintAddress(i-1);putchar('\n');
 
  return MONITOR_COMMAND;
 }
 
 if (!strncmp(chi[0].ch,"fi",2))
 {
  switch(nargs)
    {
    case 0:printf("Usage : fi ([startaddr])\n");return MONITOR_COMMAND;
    default:opcode=ValeurHexa(chi[1].ch);
    }
    
   do
     {
     printf("Enter searched bytes or string (between \"\") : ");fflush(stdout);
     gets(ch2);
     if (mon.BreakReq) {mon.BreakReq=0;printf("** Not done\n");return MONITOR_COMMAND;}   
     }
   while (!(nbbytes=ReadHexBytes(ch2,ch3)));
     
   for (i=opcode;i<=0xFFFF;i++)
     {
     for (j=0;j<nbbytes;j++) if (((unsigned char)ch3[j])!=M[i+j]) break;     
     if (j==nbbytes)  {printf("Sequence found at ");PrintAddress(i);putchar('\n');}      
     if (mon.BreakReq) {mon.BreakReq=0;break;}
     }
   
  printf("Searched from ");PrintAddress(opcode);
  printf(" to ");PrintAddress(i-1);putchar('\n');
 
  return MONITOR_COMMAND;
 }
 
 if (!strncmp(chi[0].ch,"call",4))
 {
  printf("Routine called from ");PrintAddress(Caller());putchar('\n');
  return MONITOR_COMMAND;
 }
 
 if (!strncmp(chi[0].ch,"ill",3))
 {
  mon.illeg_inst^=1;
  PrintStatus("Illegal instuctions now",mon.illeg_inst);
  return MONITOR_COMMAND;
 }
 
 if (!strncmp(chi[0].ch,"ust",3))
 {
  if (!mon.TapeSave) printf("** No save tape specified.\n");
  else {
       fclose(mon.TapeSave);
       mon.TapeSave=NULL;
       printf("Save tape updated.\n");
       ChangeTapeLoad(mon.TapeSaveName);
       }
  return MONITOR_COMMAND;
 }
 
printf("?Command unrecognized\n");
return MONITOR_COMMAND;
 
}
