/***************************************************************************/
/**                                                                       **/
/**                 The ORIC 48K emulator.UNIX version                    **/
/**                                                                       **/
/**                Written by JF FABRE (fabre@supaero.fr)                 **/
/**                                                                       **/
/**                This program may be freely distibuted                  **/
/**                                                                       **/
/**             as long as this header is included unmodified             **/
/**                                                                       **/
/***************************************************************************/

/* Low level audio controller */
/* Used for the Oric Emulator Copyright 1995 JF Fabre */

/* From a driver for HP9000 series 710/7x5 with HP-UX 9.01 */
/* Using low-level audio calls */
/* Copyright 1993 Marc Espie   (Marc.Espie@ens.fr)  */
/* Copyright 1993 Markus Gyger (mgyger@itr.ch) */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
/*#include <sys/audio.h>*/
#include <math.h>

#ifndef O_WRONLY
#define O_WRONLY      0000001
#endif

#ifndef O_NDELAY
#define O_NDELAY  0000004
#endif

#ifndef DEFAULT_SAMPLE_RATE
#define DEFAULT_SAMPLE_RATE  16000
#endif

#ifndef AUDIO_NAME
#define AUDIO_NAME           "/dev/audio"
#endif
#ifndef AUDIO_CTL_NAME
#define AUDIO_CTL_NAME       "/dev/audioCtl"
#endif

#include "protos.h"

#ifdef AUDIO_DESCRIBE

int audio_ctl;
int audio;
int sample_rate;
char *buffer;
short *sbuffer;
int index;
int channels;
int min_samples;

int stereo;
int primary, secondary;
struct audio_describe ainfo;

SND sons;

/*** setmix ***/

void set_mix(percent)
int percent;
    {
    percent *= 256;
    percent /= 100;
    primary = percent;
    secondary = 512 - percent;
    }

/*** available ***/

int available(f)
int f;
    {
    int best = 0;
    int i;

    for (i = 0; i < ainfo.nrates; i++)
	if (abs(ainfo.sample_rate[i] - f) < abs(best - f))
	    best = ainfo.sample_rate[i];
    return best;
    }

void end_all(ch)

{
printf("** %s\n",ch);
}

/** Ouvre le device audio **/

int OpenAudioDevice(f,s)
int f,s;

    {
    int type;


    audio_ctl = open(AUDIO_CTL_NAME, O_WRONLY | O_NDELAY);
    if (audio_ctl == -1)
        {
        end_all("No audio control device");return 0;
        }

    if (ioctl(audio_ctl,AUDIO_DESCRIBE, &ainfo) == -1)
        {
	end_all("No audio info");return 0;
        }

    if (f == 0)
        f = DEFAULT_SAMPLE_RATE;
        /* round frequency to acceptable value */
    sample_rate = available(f);

        /* check whether we have stereo device */
    if (ainfo.nchannels < 2)
        {
            /* a 710 or 425 -> revert to base quality audio */
        stereo = 0;
        channels = 1;
        }
    else
        {
            /* 7x5 set up */
        stereo = s;
        if (stereo)
            {
            channels = 2;
            set_mix(30);
            }
        else
            channels = 1;
        }


    if (ioctl(audio_ctl, AUDIO_SET_DATA_FORMAT, AUDIO_FORMAT_LINEAR16BIT) == -1)
        {
		  end_all("Linear format not available");return 0;
        }
    if (ioctl(audio_ctl, AUDIO_SET_CHANNELS, channels) == -1)
        {
		  end_all("Could not set # of audio channels");return 0;
        }
    if (ioctl(audio_ctl, AUDIO_SET_SAMPLE_RATE, sample_rate) == -1)
        {
		  end_all("Could not set audio sample rate");return 0;
        }

    audio = open(AUDIO_NAME, O_WRONLY | O_NDELAY);
    if (audio == -1)
        {
		  end_all("Could not open audio device");return 0;
        }
#ifdef SET_OUTPUT
        /* ensure we hear something */
    if (ioctl(audio_ctl, AUDIO_SET_OUTPUT,
        ((SET_OUTPUT & 1) ? AUDIO_OUT_INTERNAL : 0) |   /* speaker */
        ((SET_OUTPUT & 2) ? AUDIO_OUT_EXTERNAL : 0) |   /* phones  */
        ((SET_OUTPUT & 4) ? AUDIO_OUT_LINE : 0) |   /* line    */
        ((SET_OUTPUT & 7) ? 0 : AUDIO_OUT_NONE)) == -1)
		  printf("Warning: could not set audio output\n");
#endif

    index = 0;
    sbuffer = (short *)malloc(2 * channels * sample_rate);
    buffer = (char *)sbuffer;
    if (!buffer)
        {end_all("Could not allocate buffer");return 0;}
    return sample_rate;
    }
    
int update_frequency()
    {
    int oldfreq;

    oldfreq = sample_rate;
    if (ioctl(audio, AUDIO_GET_SAMPLE_RATE, &sample_rate) != -1)
        {
        if (oldfreq != sample_rate)
            {
            sbuffer = (short *)realloc(sbuffer, 2 * channels * sample_rate);
            buffer = (char *)sbuffer;
            return sample_rate;
            }
        }
    return 0;
    }

void output_samples(left, right)
int left, right;
    {
    if (stereo)
        {
        sbuffer[index++] = (left * primary + right * secondary)/256;
        sbuffer[index++] = (right * primary + left * secondary)/256;
        }
    else
        sbuffer[index++] = left + right;
    }

void flush_buffer()
    {
    if (index >= min_samples)
        {
        write(audio, buffer, 2 * index);
        index = 0;
        }
    }

void discard_buffer()
    {
    if (ioctl(audio, AUDIO_RESET, RESET_TX_BUF) == -1)
		printf("Warning: could not discard audio buffer\n");
    }

void close_audio()
    {
    free(buffer);
    close(audio);
    close(audio_ctl);
    }

/*** charger les sons de base (ZAP,PING,...) ***/

void ChargerSons()

{
if (!LoadSoundFile("sounds/zap.l16",&(sons.son_zap))) printf("** Zap sound not loaded\n");
if (!LoadSoundFile("sounds/ping.l16",&(sons.son_ping))) printf("** Ping sound not loaded\n");
if (!LoadSoundFile("sounds/keybeep.l16",&(sons.son_beep))) printf("** Beep sound not loaded\n");
if (!LoadSoundFile("sounds/contbeep.l16",&(sons.son_contbeep))) printf("** ContBeep sound not loaded\n");
if (!LoadSoundFile("sounds/explode.l16",&(sons.son_explode))) printf("** Explode sound not loaded\n");
if (!LoadSoundFile("sounds/shoot.l16",&(sons.son_shoot))) printf("** Shoot sound not loaded\n");
}

/*** charger un fichier son en memoire ***/

unsigned int LoadSoundFile(fname,pson)

char *fname;
son *pson;
{
FILE *f;
long i;

if (!(f=fopen(fname,"rb"))) return 0;

fseek(f,0,SEEK_END);
pson->taille=ftell(f); 
rewind(f);

pson->buff=(unsigned char *)malloc(pson->taille*sizeof(unsigned char));
if (pson->buff==NULL) return 0;

fread(pson->buff,pson->taille,1,f);

fclose(f);

  
return 1;

}

/*** faire jouer le son ***/

void PlaySound(un_son)

son un_son;

{
write(audio,un_son.buff,un_son.taille);
sync();
}

/*** Initialiser le serveur de sons ***/

int InitAudioProcess(comm_struct *pcs)

{
char ch[10];

if (pipe(pcs->fd_pere)==-1)  /** ouverture d'un pipe **/
      return 0;
if (pipe(pcs->fd_fils)==-1)  /** ouverture d'un autre pipe **/
      return 0;

/*fcntl(audio,F_SETFL,fcntl(audio,F_GETFL,0)|O_NONBLOCK);*/  

if (pcs->id=fork())       /** duplication du processus en 2 processus **/
                          /** pere et fils **/
                          
/**** phase executee par le pere uniquement ****/
   {
   close(pcs->fd_pere[0]);  /** fermeture des canaux pipe **/
   close(pcs->fd_fils[1]);  /** inutiles pour le pere **/

   return 1;    
   }
/**** phase executee par le fils uniquement ****/

   else
   {  
   
    
   close(pcs->fd_fils[0]);   /** fermeture des canaux pipe **/
   close(pcs->fd_pere[1]);   /** inutiles pour le fils **/
   ChargerSons();


   for (;;)
   {   
     read(pcs->fd_pere[0],ch,1); /** attente bloquante **/
     switch(ch[0])
       {
       case 'Z': PlaySound(sons.son_zap);break;
       case 'P': PlaySound(sons.son_ping);break;
       case 'S': PlaySound(sons.son_shoot);break;
       case 'E': PlaySound(sons.son_explode);break;
       case 'C': PlaySound(sons.son_contbeep);break;
       case 'K': PlaySound(sons.son_beep);break;
       default : printf("** Audio stopped.\n");exit(0);break;
       }
    }
   }

}

#endif

